// exhappyend.cpp, v1.0 2013/10/04
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts GPDA64BY (*.dat) archives used by 
// Ore no Imouto ga Konna ni Kawaii Wake ga Nai: Happy End (PS3).

#include "as-util.h"
#include "as-ps3.h"
#include "as-zlib.h"
#include "zlib.h"

struct DATHDR {
  uint8_t  signature[8]; // "GPDA64BY"
  uint64_t archive_size;
  uint32_t unknown1;
  uint32_t entry_count;
};

struct DATENTRY {
  uint64_t offset;
  uint64_t is_compressed;
  uint64_t length;
  uint64_t filename_offset;
};

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exhappyend v1.0, coded by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.dat>\n", argv[0]);
    return -1;
  }

  string in_filename = argv[1];

  int fd = as::open_or_die(in_filename, O_RDONLY | O_BINARY); 

  DATHDR hdr;
  read(fd, &hdr, sizeof(hdr));

  auto entries = new DATENTRY[hdr.entry_count];
  read(fd, entries, sizeof(DATENTRY) * hdr.entry_count);

  for (uint32_t i = 0; i < hdr.entry_count; i++) {
    uint32_t filename_len = 0;
    _lseeki64(fd, entries[i].filename_offset, SEEK_SET);
    read(fd, &filename_len, sizeof(filename_len));

    char filename_buff[1024] = { 0 };
    read(fd, filename_buff, filename_len);
    string filename(filename_buff, filename_len);

    uint32_t len  = (uint32_t) entries[i].length;
    auto     buff = new uint8_t[len];
    _lseeki64(fd, entries[i].offset, SEEK_SET);
    read(fd, buff, len);

    if (entries[i].is_compressed) {
      as::inflate(buff, 
                  len, 
                  1024 * 1024 * 20); // "big enough"
    }

    as::write_file(filename, buff, len);

    // The data is a flabby tree of serialized objects I don't feel like
    // handling.  This laaaaame hack suffices for now.
    if (!memcmp(buff, "PHYR", 4)) {
      as::replace_substr(filename, ".phyre", "");

      auto end = buff + len;

      for (auto p = buff; p + 5 < end; p++) {
        if (!memcmp(p, "DXT5", 4) || !memcmp(p, "DXT1", 4)) {
          uint32_t width    = as::flip_endian(*(uint32_t*) (p - 143));
          uint32_t height   = as::flip_endian(*(uint32_t*) (p - 139));
          uint32_t dds_len  = as::flip_endian(*(uint32_t*) (buff + 72));
          auto     dds_buff = p + 42;

          struct DDSHDR {
            uint8_t  signature[4]; // "DDS "
            uint32_t unknown1;
            uint32_t unknown2;
            uint32_t height;
            uint32_t width;
            uint32_t length;
            uint32_t unknown3[15];
            uint8_t  type[4]; // "DXT1" etc
            uint32_t depth;
            uint32_t unknown4[9];
          };

          DDSHDR ddshdr = {
            { 0x44, 0x44, 0x53, 0x20 }, 0x0000007C, 0x000A1007, 0x00000000, 
            0x00000000, 0x00000000, 0x00000000, 0x00000001, 
            0x00000000, 0x00000000, 0x00000000, 0x00000000, 
            0x00000000, 0x00000000, 0x00000000, 0x00000000, 
            0x00000000, 0x00000000, 0x00000000, 0x00000020, 
            0x00000005, { 0 },      0x00000000, 0x00000000, 
            0x00000000, 0x00000000, 0x00000000, 0x00401008, 
            0x00000000, 0x00000000, 0x00000000, 0x00000000, 
          };

          memcpy(ddshdr.type, p, 4);
          ddshdr.width = width;
          ddshdr.height = height;

          int dds_fd = as::open_or_die(as::get_file_prefix(filename) + ".dds",
                                   O_CREAT | O_TRUNC | O_WRONLY | O_BINARY,
                                   S_IREAD | S_IWRITE);

          write(dds_fd, &ddshdr, sizeof(ddshdr));
          write(dds_fd, dds_buff, dds_len);

          close(dds_fd);

          break;
        }

        if (!memcmp(p, "ARGB8", 5)) {
          uint32_t width     = as::flip_endian(*(uint32_t*) (p - 143));
          uint32_t height    = as::flip_endian(*(uint32_t*) (p - 139));
          uint32_t flags     = as::flip_endian(*(uint32_t*) (p - 91));
          auto     rgba_buff = p + 43;

          uint32_t rgba_len  = width * height * 4;

          if (rgba_buff + rgba_len <= end) {
            as::image_t image(width, height, 4, rgba_buff);

            // Format? Flags? nfi
            if (flags == 0x18529) {   
              as::unswizzle_ps3(image.buffer(), image.length(), width, height);
              image.flip();
            }

            image.write(as::get_file_prefix(filename) + ".bmp");
          }

          break;
        }
      }
    }

    delete [] buff;
  }

  delete [] entries;

  close(fd);

  return 0;
}
